
#include <gtk/gtk.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <signal.h>
#include <unistd.h>

#include "xmms/plugin.h"
#include "libxmms/configfile.h"

static void init(void);
static void cleanup(void);
static void configure(void);
static gint timeout_func(gpointer);
static gint timeout_tag = 0;
static gint previous_song = -1, previous_length = -2;
static gchar *cmd_line = NULL;

static GtkWidget *configure_win = NULL, *configure_vbox;
static GtkWidget *cmd_hbox, *cmd_label, *cmd_entry;
static GtkWidget *configure_separator;
static GtkWidget *configure_bbox, *configure_ok, *configure_cancel;

GeneralPlugin sc_gp =
{
	NULL,			/* handle */
	NULL,			/* filename */
	-1,			/* xmms_session */
	"Song Change",
	init,
	NULL,
	configure,
	cleanup,
};

GeneralPlugin *get_gplugin_info(void)
{
	return &sc_gp;
}

static void read_config(void)
{
	ConfigFile *cfgfile;
	gchar *filename;

	if (cmd_line)
		g_free(cmd_line);
	cmd_line = NULL;

	filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);
	cfgfile = xmms_cfg_open_file(filename);
	if (cfgfile)
	{
		xmms_cfg_read_string(cfgfile, "song_change", "cmd_line", &cmd_line);
		xmms_cfg_free(cfgfile);
	}
	g_free(filename);
}

static void init(void)
{
	read_config();

	previous_song = -1;
	previous_length = -2;
	timeout_tag = gtk_timeout_add(100, timeout_func, NULL);
}

static void cleanup(void)
{
	if (timeout_tag)
		gtk_timeout_remove(timeout_tag);
	timeout_tag = 0;
	if (cmd_line)
		g_free(cmd_line);
	cmd_line = NULL;
}

static void configure_ok_cb(gpointer data)
{
	gchar *cmd, *filename;
	ConfigFile *cfgfile;

	cmd = gtk_entry_get_text(GTK_ENTRY(cmd_entry));
	filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);

	cfgfile = xmms_cfg_open_file(filename);
	if (!cfgfile)
		cfgfile = xmms_cfg_new();
	xmms_cfg_write_string(cfgfile, "song_change", "cmd_line", cmd);
	xmms_cfg_write_file(cfgfile, filename);
	xmms_cfg_free(cfgfile);
	g_free(filename);
	if (timeout_tag)
	{
		if (cmd_line)
			g_free(cmd_line);
		cmd_line = g_strdup(cmd);
	}
	gtk_widget_destroy(configure_win);
}

static void configure(void)
{
	if (!configure_win)
	{
		read_config();

		configure_win = gtk_window_new(GTK_WINDOW_DIALOG);
		gtk_signal_connect(GTK_OBJECT(configure_win), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &configure_win);
		gtk_window_set_title(GTK_WINDOW(configure_win), "Song Change Configuration");
		gtk_window_set_position(GTK_WINDOW(configure_win), GTK_WIN_POS_MOUSE);

		gtk_container_set_border_width(GTK_CONTAINER(configure_win), 10);

		configure_vbox = gtk_vbox_new(FALSE, 10);
		gtk_container_add(GTK_CONTAINER(configure_win), configure_vbox);

		cmd_hbox = gtk_hbox_new(FALSE, 5);
		gtk_box_pack_start(GTK_BOX(configure_vbox), cmd_hbox, FALSE, FALSE, 0);

		cmd_label = gtk_label_new("Command:");
		gtk_box_pack_start(GTK_BOX(cmd_hbox), cmd_label, FALSE, FALSE, 0);
		gtk_widget_show(cmd_label);

		cmd_entry = gtk_entry_new();
		if (cmd_line)
			gtk_entry_set_text(GTK_ENTRY(cmd_entry), cmd_line);
		gtk_widget_set_usize(cmd_entry, 200, -1);
		gtk_box_pack_start(GTK_BOX(cmd_hbox), cmd_entry, TRUE, TRUE, 0);
		gtk_widget_show(cmd_entry);
		gtk_widget_show(cmd_hbox);

		configure_separator = gtk_hseparator_new();
		gtk_box_pack_start(GTK_BOX(configure_vbox), configure_separator, FALSE, FALSE, 0);
		gtk_widget_show(configure_separator);

		configure_bbox = gtk_hbutton_box_new();
		gtk_button_box_set_layout(GTK_BUTTON_BOX(configure_bbox), GTK_BUTTONBOX_END);
		gtk_button_box_set_spacing(GTK_BUTTON_BOX(configure_bbox), 5);
		gtk_box_pack_start(GTK_BOX(configure_vbox), configure_bbox, FALSE, FALSE, 0);

		configure_ok = gtk_button_new_with_label("Ok");
		gtk_signal_connect(GTK_OBJECT(configure_ok), "clicked", GTK_SIGNAL_FUNC(configure_ok_cb), NULL);
		GTK_WIDGET_SET_FLAGS(configure_ok, GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(configure_bbox), configure_ok, TRUE, TRUE, 0);
		gtk_widget_show(configure_ok);
		gtk_widget_grab_default(configure_ok);

		configure_cancel = gtk_button_new_with_label("Cancel");
		gtk_signal_connect_object(GTK_OBJECT(configure_cancel), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(configure_win));
		GTK_WIDGET_SET_FLAGS(configure_cancel, GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(configure_bbox), configure_cancel, TRUE, TRUE, 0);
		gtk_widget_show(configure_cancel);
		gtk_widget_show(configure_bbox);
		gtk_widget_show(configure_vbox);
		gtk_widget_show(configure_win);
	}
}

static void bury_child(int signal)
{
	waitpid(-1, NULL, WNOHANG);
}

static gint timeout_func(gpointer data)
{
	gint pos, length, i;
	gchar *cmd, *argv[4], *titlestr, *shstring;

	GDK_THREADS_ENTER();

	pos = xmms_remote_get_playlist_pos(sc_gp.xmms_session);
	length = xmms_remote_get_playlist_time(sc_gp.xmms_session, pos);
	if (pos != previous_song || length != previous_length)
	{
		previous_song = pos;
		previous_length = length;
		if (cmd_line)
		{
			signal(SIGCHLD, bury_child);
			if (fork() == 0)
			{
				/* We don't want this process to hog the audio device etc */
				for (i = 3; i < 255; i++)
					close(i);

				titlestr = g_strdup_printf("\"%s\"", xmms_remote_get_playlist_title(sc_gp.xmms_session, pos));
				shstring = g_strdup_printf(cmd_line, titlestr);
				g_free(titlestr);

				argv[0] = "/bin/sh";
				argv[1] = "-c";
				argv[2] = shstring;
				argv[3] = NULL;
				execv("/bin/sh", argv);

				g_free(shstring);
			}
		}
	}

	GDK_THREADS_LEAVE();

	return TRUE;
}
