/*  XMMS - Cross-platform multimedia player
 *  Copyright (C) 1998-1999  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "OSS.h"
#include "libxmms/util.h"
#include <errno.h>

#define NFRAGS		32

#define min(x,y) ((x)<(y)?(x):(y))

static gint fd = 0;
static gpointer buffer;
static gboolean going = FALSE, prebuffer, paused = FALSE, unpause = FALSE,
         do_pause = FALSE, remove_prebuffer = FALSE;
static gint buffer_size, prebuffer_size, blk_size;
static gint rd_index = 0, wr_index = 0;
static gint output_time_offset = 0, written = 0, output_bytes = 0;
static gint bps, ebps;
static gint flush;
static gint fragsize, oss_format, format, channels, frequency, efrequency, device_buffer_size;
static gint input_bps, input_format, input_frequency, input_channels;
static gchar *device_name;
static pthread_t buffer_thread;
static gboolean realtime = FALSE;

static void oss_setup_format(AFormat fmt,gint rate, gint nch)
{
	format = fmt;
	frequency = rate;
	channels = nch;
	switch (fmt)
	{
		case FMT_U8:
			oss_format = AFMT_U8;
			break;
		case FMT_S8:
			oss_format = AFMT_S8;
			break;
		case FMT_U16_LE:
			oss_format = AFMT_U16_LE;
			break;
		case FMT_U16_BE:
			oss_format = AFMT_U16_BE;
			break;
		case FMT_U16_NE:
#ifdef AFMT_U16_NE
			oss_format = AFMT_U16_NE;
#else
#ifdef WORDS_BIGENDIAN
			oss_format = AFMT_U16_BE;
#else
			oss_format = AFMT_U16_LE;
#endif
#endif
			break;
		case FMT_S16_LE:
			oss_format = AFMT_S16_LE;
			break;
		case FMT_S16_BE:
			oss_format = AFMT_S16_BE;
			break;
		case FMT_S16_NE:
#ifdef AFMT_S16_NE
			oss_format = AFMT_S16_NE;
#else
#ifdef WORDS_BIGENDIAN
			oss_format = AFMT_S16_BE;
#else
			oss_format = AFMT_S16_LE;
#endif
#endif
			break;
	}

	bps = rate * nch;
	if (oss_format == AFMT_U16_BE || oss_format == AFMT_U16_LE || oss_format == AFMT_S16_BE || oss_format == AFMT_S16_LE)
		bps *= 2;
	fragsize = 0;
	while ((1L << fragsize) < bps / 25)
		fragsize++;
	fragsize--;

	device_buffer_size = ((1L << fragsize) * (NFRAGS + 1));
}
	

gint oss_get_written_time(void)
{
	if (!going)
		return 0;
	return (gint) (((gfloat) written * 1000) / (gfloat) (input_bps));
}

gint oss_get_output_time(void)
{
	audio_buf_info buf_info;
	gint bytes;

	if (!fd || !going)
		return 0;

	if (!paused)
	{
		if (!ioctl(fd, SNDCTL_DSP_GETOSPACE, &buf_info))
			bytes = output_bytes - ((buf_info.fragstotal - buf_info.fragments) * buf_info.fragsize);
		else
			bytes = output_bytes;
	}
	else
		bytes = output_bytes;

	if (bytes < 0)
		bytes = 0;
	return output_time_offset + (gint) ((gfloat) ((bytes) * 1000.0) / (gfloat) ebps);
}

gint oss_used(void)
{
	if (realtime)
		return 0;
	else
	{
		if (wr_index >= rd_index)
			return wr_index - rd_index;
		return buffer_size - (rd_index - wr_index);
	}
}

gint oss_playing(void)
{
	audio_buf_info buf_info;
	int bytes;

	if (!ioctl(fd, SNDCTL_DSP_GETOSPACE, &buf_info))
		bytes = ((buf_info.fragstotal - buf_info.fragments - 3) * buf_info.fragsize);
	else
		bytes = 0;

	if (!oss_used() && bytes <= 0)
		return FALSE;

	return TRUE;
}

gint oss_free(void)
{
	if (!realtime)
	{
		if (remove_prebuffer && prebuffer)
		{
			prebuffer = FALSE;
			remove_prebuffer = FALSE;
		}
		if (prebuffer)
			remove_prebuffer = TRUE;

		if (rd_index > wr_index)
			return (rd_index - wr_index) - device_buffer_size - 1;
		return (buffer_size - (wr_index - rd_index)) - device_buffer_size - 1;
	}
	else
		if (paused)
			return 0;
		else
			return 1000000;
}

static void oss_write_audio(gpointer data,gint length)
{
	AFormat new_format;
	gint new_frequency,new_channels;
	EffectPlugin *ep;
	audio_buf_info abuf_info;
	
	new_format = input_format;
	new_frequency = input_frequency;
	new_channels = input_channels;
	
	ep = get_current_effect_plugin();
	if(effects_enabled() && ep && ep->query_format)
	{
		ep->query_format(&new_format,&new_frequency,&new_channels);
	}
	
	if(new_format != format || new_frequency != frequency || new_channels != channels)
	{
		output_time_offset += (gint) ((gfloat) ((output_bytes) * 1000.0) / (gfloat) ebps);
		output_bytes = 0;
		oss_setup_format(new_format, new_frequency, new_channels);
		frequency = new_frequency;
		channels = new_channels;
		close(fd);
		fd = open(device_name,O_WRONLY);
		oss_set_audio_params();
	}
	if(effects_enabled() && ep && ep->mod_samples)
		length = ep->mod_samples(&data,length, input_format, input_frequency, input_channels);
	ioctl(fd,SNDCTL_DSP_GETOSPACE,&abuf_info);
	while(abuf_info.bytes < length)
	{
		xmms_usleep(10000);
		ioctl(fd,SNDCTL_DSP_GETOSPACE,&abuf_info);
	}
	if(frequency == efrequency)
		output_bytes += write(fd,data,length);
	else
		output_bytes += oss_downsample(data,length,frequency,efrequency);
}

gint oss_downsample(guchar * ob, guint length, guint speed, guint espeed)
{
	guint nlen, i, off, d, w;

	if ((oss_format == AFMT_U16_BE || oss_format == AFMT_U16_LE || oss_format == AFMT_S16_BE || oss_format == AFMT_S16_LE) && channels == 2)
	{
		gulong *nbuffer, *obuffer, *ptr;

		obuffer = (gulong *) ob;
		length >>= 2;

		nlen = (length * espeed) / speed;
		d = (speed << 8) / espeed;

		nbuffer = g_malloc(nlen << 2);
		for (i = 0, off = 0, ptr = nbuffer; i < nlen; i++)
		{
			*ptr++ = obuffer[off >> 8];
			off += d;
		}
		w = write(fd, nbuffer, nlen << 2);
		g_free(nbuffer);
	}
	else if (((oss_format == AFMT_U16_BE || oss_format == AFMT_U16_LE || oss_format == AFMT_S16_BE || oss_format == AFMT_S16_LE) && channels == 1)
	     || ((oss_format == AFMT_U8 || oss_format == AFMT_S8) && channels == 2))
	{
		gushort *nbuffer, *obuffer, *ptr;

		obuffer = (gushort *) ob;
		length >>= 1;

		nlen = (length * espeed) / speed;
		d = (speed << 8) / espeed;

		nbuffer = g_malloc(nlen << 1);
		for (i = 0, off = 0, ptr = nbuffer; i < nlen; i++)
		{
			*ptr++ = obuffer[off >> 8];
			off += d;
		}
		w = write(fd, nbuffer, nlen << 1);
		g_free(nbuffer);
	}
	else
	{
		guchar *nbuffer, *obuffer, *ptr;

		obuffer = ob;

		nlen = (length * espeed) / speed;
		d = (speed << 8) / espeed;

		nbuffer = g_malloc(nlen);
		for (i = 0, off = 0, ptr = nbuffer; i < nlen; i++)
		{
			*ptr++ = obuffer[off >> 8];
			off += d;
		}
		w = write(fd, nbuffer, nlen);
		g_free(nbuffer);
	}
	return w;

}

void oss_write(gpointer ptr, gint length)
{
	gint cnt, off = 0;
	audio_buf_info abuf_info;

	if (!realtime)
	{
		remove_prebuffer = FALSE;

		written += length;
		while (length > 0)
		{
			cnt = min(length, buffer_size - wr_index);
			memcpy(buffer + wr_index, ptr + off, cnt);
			wr_index = (wr_index + cnt) % buffer_size;
			length -= cnt;
			off += cnt;

		}
	}
	else
	{
		if (paused)
			return;
		oss_write_audio(ptr,length);
		written += length;
		
	}

}

void oss_close(void)
{
	going = 0;
	if (!realtime)
		pthread_join(buffer_thread, NULL);
	else
	{
		ioctl(fd, SNDCTL_DSP_RESET, 0);
		close(fd);
	}
	wr_index = 0;
	rd_index = 0;
}

void oss_flush(gint time)
{
	if (!realtime)
	{
		flush = time;
		while (flush != -1)
			xmms_usleep(10000);
	}
	else
	{
		ioctl(fd, SNDCTL_DSP_RESET, 0);
		close(fd);
		fd = open(device_name, O_WRONLY);
		oss_set_audio_params();
		output_time_offset = time;
		written = (time / 10) * (input_bps / 100);
		output_bytes = 0;
	}
}

void oss_pause(short p)
{
	if (!realtime)
	{
		if (p == TRUE)
			do_pause = TRUE;
		else
			unpause = TRUE;
	}
	else
		paused = p;

}

void *oss_loop(void *arg)
{
	gint length, cnt;
	audio_buf_info abuf_info;
	

	while (going)
	{
		if (oss_used() > prebuffer_size)
			prebuffer = FALSE;
		if (oss_used() > 0 && !paused && !prebuffer)
		{
			length = min(blk_size, oss_used());
			while (length > 0)
			{
				cnt = min(length,buffer_size-rd_index);
				oss_write_audio(buffer + rd_index, cnt);
				rd_index=(rd_index+cnt)%buffer_size;
				length-=cnt;				
			}
			if (!oss_used())
				ioctl(fd, SNDCTL_DSP_POST, 0);
		}
		else
			xmms_usleep(10000);
		if (do_pause && !paused)
		{
			do_pause = FALSE;
			paused = TRUE;
			if (!ioctl(fd, SNDCTL_DSP_GETOSPACE, &abuf_info))
			{
				rd_index -= (abuf_info.fragstotal - abuf_info.fragments) * abuf_info.fragsize;
				output_bytes -= (abuf_info.fragstotal - abuf_info.fragments) * abuf_info.fragsize;
			}
			if (rd_index < 0)
				rd_index += buffer_size;
			ioctl(fd, SNDCTL_DSP_RESET, 0);

		}
		if (unpause && paused)
		{
			unpause = FALSE;
			close(fd);
			fd = open(device_name, O_WRONLY);
			oss_set_audio_params();
			paused = FALSE;
		}

		if (flush != -1)
		{
			/*
			 * This close and open is a work around of a bug that exists in some drivers which 
			 * cause the driver to get fucked up by a reset
			 */

			ioctl(fd, SNDCTL_DSP_RESET, 0);
			close(fd);
			fd = open(device_name, O_WRONLY);
			oss_set_audio_params();
			output_time_offset = flush;
			written = (flush / 10) * (input_bps / 100);
			rd_index = wr_index = output_bytes = 0;
			flush = -1;
			prebuffer = TRUE;
		}

	}

	ioctl(fd, SNDCTL_DSP_RESET, 0);
	close(fd);
	g_free(buffer);
	pthread_exit(NULL);
}

void oss_set_audio_params(void)
{
	gint frag, stereo;

	ioctl(fd, SNDCTL_DSP_RESET, 0);
	frag = (NFRAGS << 16) | fragsize;
	ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &frag);
	ioctl(fd, SNDCTL_DSP_SETFMT, &oss_format);
	ioctl(fd, SNDCTL_DSP_SETFMT, &oss_format);
	stereo = channels - 1;
	ioctl(fd, SNDCTL_DSP_STEREO, &stereo);
	efrequency = frequency;
	ioctl(fd, SNDCTL_DSP_SPEED, &efrequency);
	if(abs(((efrequency * 100) / frequency) - 100) < 10)
		efrequency = frequency;
	if(ioctl(fd, SNDCTL_DSP_GETBLKSIZE, &blk_size) == -1)
		blk_size = 1L << fragsize;

	ebps = efrequency * channels;
	if (oss_format == AFMT_U16_BE || oss_format == AFMT_U16_LE || oss_format == AFMT_S16_BE || oss_format == AFMT_S16_LE)
		ebps *= 2;
}

gint oss_open(AFormat fmt, gint rate, gint nch)
{
	gint fmts, policy;
	
	oss_setup_format(fmt,rate,nch);
	
	input_format = format;
	input_channels = channels;
	input_frequency = frequency;
	input_bps = bps;

	realtime = xmms_check_realtime_priority();
	
	if(!realtime)
	{
		buffer_size = (oss_cfg.buffer_size * input_bps) / 1000;
		if (buffer_size < 8192)
			buffer_size = 8192;
		prebuffer_size = (buffer_size * oss_cfg.prebuffer) / 100;
		if (buffer_size - prebuffer_size < 4096)
			prebuffer_size = buffer_size - 4096;

		buffer_size += device_buffer_size;
		buffer = g_malloc0(buffer_size);
	}
	going = 1;
	flush = -1;
	prebuffer = 1;
	wr_index = rd_index = output_time_offset = written = output_bytes = 0;
	paused = FALSE;
	do_pause = FALSE;
	unpause = FALSE;
	remove_prebuffer = FALSE;

	if (oss_cfg.audio_device > 0)
		device_name = g_strdup_printf("/dev/dsp%d", oss_cfg.audio_device);
	else
		device_name = g_strdup("/dev/dsp");
	fd = open(device_name, O_WRONLY);
	if (fd == -1)
	{
		g_free(buffer);
		return 0;
	}
	oss_set_audio_params();
	if (!realtime)
		pthread_create(&buffer_thread, NULL, oss_loop, NULL);
	return 1;
}
