/*  XMMS - Cross-platform multimedia player
 *  Copyright (C) 1998-1999  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "xmms.h"
#include <time.h>
#include "libxmms/util.h"


GList *playlist = NULL;
GList *shuffle_list = NULL;
static gboolean playlist_get_info_scan_active = FALSE, playlist_get_info_going = FALSE;
static pthread_t playlist_get_info_thread;
pthread_mutex_t playlist_mutex = PTHREAD_MUTEX_INITIALIZER;
PlaylistEntry *playlist_position;

extern PlayList_List *playlistwin_list;
extern Vis *mainwin_vis;
extern SVis *mainwin_svis;

static GList *find_playlist_position_list(void)
{
	if (cfg.shuffle)
		return g_list_find(shuffle_list, playlist_position);
	return g_list_find(playlist, playlist_position);
}

void playlist_clear(void)
{
	GList *node;
	PlaylistEntry *entry;

	pthread_mutex_lock(&playlist_mutex);

	if (playlist)
	{
		if (get_input_playing())
			input_stop();
		node = playlist;
		while (node)
		{
			entry = (PlaylistEntry *) node->data;
			if (entry->filename)
				g_free(entry->filename);
			if (entry->title)
				g_free(entry->title);
			g_free(entry);
			node = node->next;
		}
		g_list_free(playlist);
		playlist = NULL;
		playlist_position = NULL;
	}

	pthread_mutex_unlock(&playlist_mutex);
	playlist_generate_shuffle_list();
}

void playlist_delete(gboolean crop)
{
	gboolean restart_playing = FALSE, set_info_text = FALSE, regen_shuffle = FALSE;
	GList *node = playlist, *next;
	PlaylistEntry *entry;
	GList *plist_pos_list;

	pthread_mutex_lock(&playlist_mutex);
	while (node)
	{
		entry = (PlaylistEntry *) node->data;
		next = g_list_next(node);
		if ((entry->selected && !crop) || (!entry->selected && crop))
		{
			regen_shuffle = TRUE;
			/*
			 * We call g_list_find manually here because we don't want an item in the shuffle_list
			 */
			plist_pos_list = g_list_find(playlist, playlist_position);
			if (plist_pos_list == node)
			{
				set_info_text = TRUE;
				if (get_input_playing())
				{
					pthread_mutex_unlock(&playlist_mutex);
					input_stop();
					pthread_mutex_lock(&playlist_mutex);
					if(g_list_index(playlist, entry) == -1)
					{
						node = playlist;
						continue;
					}
					restart_playing = TRUE;
				}
				if (g_list_next(plist_pos_list))
					playlist_position = g_list_next(plist_pos_list)->data;
				else if (g_list_previous(plist_pos_list))
					playlist_position = g_list_previous(plist_pos_list)->data;
				else if (node != playlist)
					playlist_position = playlist->data;
				else
					playlist_position = NULL;
			}
			if (entry->filename)
				g_free(entry->filename);
			if (entry->title)
				g_free(entry->title);
			playlist = g_list_remove_link(playlist, node);
			g_free(entry);
			g_list_free_1(node);
		}
		node = next;
	}
	pthread_mutex_unlock(&playlist_mutex);
	
	playlistwin_update_list();
	if (regen_shuffle)
		playlist_generate_shuffle_list();
	if (set_info_text)
		mainwin_set_info_text();
	if (restart_playing)
	{
		if (playlist_position)
			playlist_play();
		else
			mainwin_set_song_info(0, 0, 0);
	}

}

static void __playlist_ins(gchar * filename, glong pos)
{
	PlaylistEntry *entry;

	entry = g_malloc0(sizeof (PlaylistEntry));
	entry->filename = g_strdup(filename);
	entry->length = -1;

	pthread_mutex_lock(&playlist_mutex);
	playlist = g_list_insert(playlist, entry, pos);
	pthread_mutex_unlock(&playlist_mutex);		

	playlist_get_info_scan_active = TRUE;
}

void playlist_ins(gchar * filename, glong pos)
{
	gchar *ext;
	
	ext = strrchr(filename, '.');
	if (ext && ((!strcasecmp(ext, ".m3u") || !strcasecmp(ext, ".pls"))))
		playlist_load_ins(filename, pos);
	else
		if (input_check_file(filename))
			__playlist_ins(filename, pos);
}

/*  void playlist_add_dir(gchar * path) */
/*  { */
/*  	playlist_ins_dir(path, -1); */
/*  } */

guint playlist_ins_dir(gchar * path, glong pos)
{
	DIR *dir;
	struct dirent *dirent;
	struct stat statbuf;
	gchar *filename, *temp;
	guint entries = 0;
	GList *list, *node;

	if (path[strlen(path) - 1] != '/')
		temp = g_strdup_printf("%s/", path);
	else
		temp = g_strdup(path);

	if ((list = input_scan_dir(temp)) != NULL)
	{
		g_free(temp);
		node = list;
		while (node)
		{
			__playlist_ins(node->data, pos);
			entries++;
			if (pos > 0)
				pos++;
			g_free(node->data);
			node = g_list_next(node);
		}
		g_list_free(list);
		playlist_generate_shuffle_list();
		return entries;
	}
	g_free(temp);
	if ((dir = opendir(path)) != NULL)
	{
		while ((dirent = readdir(dir)) != NULL)
		{
			if (dirent->d_name[0] != '.')
			{
				filename = g_strconcat(path, "/", dirent->d_name, NULL);
				lstat(filename, &statbuf);
				if (S_ISDIR(statbuf.st_mode))
				{
					guint i;
					i = playlist_ins_dir(filename, pos);
					entries += i;
					if (pos > 0)
						pos += i;
				}
				else if (input_check_file(filename))
				{
					__playlist_ins(filename, pos);
					entries++;
					if (pos > 0)
						pos++;
				}
				g_free(filename);
			}
		}
		playlist_generate_shuffle_list();
		closedir(dir);
	}
	return entries;
}

/*
 * I'm not sure if this is the correct way to do it but it seems to work fine
 */

/*  void playlist_add_url_string(gchar * string) */
/*  { */
/*  	playlist_ins_url_string(string, -1); */
/*  } */

guint playlist_ins_url_string(gchar * string, glong pos)
{
	gchar *temp, *ext;
	struct stat statbuf;
	gint i = 1, entries = 0;

	while (*string)
	{
		temp = strchr(string, '\n');
		if (temp)
		{
			if (*(temp - 1) == '\r')
				*(temp - 1) = '\0';
			*temp = '\0';
		}
		if (!strncasecmp(string, "file:", 5))
			string = string + 5;

		stat(string, &statbuf);
		if (S_ISDIR(statbuf.st_mode))
			i = playlist_ins_dir(string, pos);
		else
		{
			ext = strrchr(string, '.');
			if (ext && (!strcasecmp(ext, ".m3u") || !strcasecmp(ext, ".pls")))
				i = playlist_load_ins(string, pos);
			else
			{
				playlist_ins(string, pos);
				i = 1;
			}
		}
		entries += i;
		if (pos > 0)
			pos += i;
		if (!temp)
			break;
		string = temp + 1;
	}
	playlist_generate_shuffle_list();

	return entries;
}

void playlist_play(void)
{
	PlaylistEntry *entry;

	if (get_playlist_length() == 0)
		return;

	/* If the user wants a skin randomly selected on play */
	if (cfg.random_skin_on_play)
	{
		/* Get the current skin list */
		scan_skins();
		/* If there are entries */
		if (g_list_length(skinlist)) {
			/* Get a random value to select the skin to use */
			int randval = (int)(rand() % (g_list_length(skinlist) + 1));
			/* If the random value is 0, use the default skin */
			/* Otherwise subtract 1 from the random value and */
			/* select the skin */
			load_skin(randval == 0 ? NULL : ((struct SkinNode *) g_list_nth(skinlist, randval - 1)->data)->path);
		}
		/* Call scan_skins() again to make sure skin selector */
		/* is up to date */
		scan_skins();
	}

	if (get_input_playing())
		input_stop();

	vis_clear_data(mainwin_vis);
	vis_clear_data(playlistwin_vis);
	svis_clear_data(mainwin_svis);

	PL_LOCK();
	entry = g_list_nth(playlist, get_playlist_position())->data;
	PL_UNLOCK();
	
	input_play(entry->filename);

	if (input_get_time() != -1)
	{
		equalizerwin_load_auto_preset(entry->filename);
		input_set_eq(cfg.equalizer_active, cfg.equalizer_preamp, cfg.equalizer_bands);
	}
	playlist_check_pos_current();
}

void playlist_set_info(gchar * title, gint length, gint rate, gint freq, gint nch)
{

	PlaylistEntry *entry;

	PL_LOCK();
	entry = g_list_nth(playlist, get_playlist_position())->data;

	if (entry && entry->title)
	{
		g_free(entry->title);
		entry->title = NULL;
	}
	if (entry && title)
		entry->title = g_strdup(title);
	entry->length = length;
	PL_UNLOCK();
	mainwin_set_song_info(rate, freq, nch);
	mainwin_set_info_text();
	playlistwin_update_list();
}

void playlist_check_pos_current(void)
{
	guint pos;
	
	pos = g_list_index(get_playlist(), playlist_position);

	if (playlistwin_list && !playlistwin_item_visible(pos))
	{
		if (pos < playlistwin_list->pl_num_visible / 2)
			playlistwin_set_toprow(0);
		else if (get_playlist_length() - pos < playlistwin_list->pl_num_visible / 2)
			playlistwin_set_toprow(get_playlist_length() - playlistwin_list->pl_num_visible);
		else
			playlistwin_set_toprow(pos - (playlistwin_list->pl_num_visible / 2));
	}
}

void playlist_next(void)
{
	GList *plist_pos_list;

	PL_LOCK();
	if (!playlist)
	{
		PL_UNLOCK();
		return;
	}

	plist_pos_list = find_playlist_position_list();

	if (!cfg.repeat && !g_list_next(plist_pos_list))
	{
		PL_UNLOCK();
		return;
	}

	if (g_list_next(plist_pos_list))
		playlist_position = g_list_next(plist_pos_list)->data;
	else if (cfg.repeat)
	{
		if (cfg.shuffle)
			playlist_position = shuffle_list->data;
		else
			playlist_position = playlist->data;
	}
	PL_UNLOCK();
	playlist_check_pos_current();
	if (get_input_playing())
	{
		input_stop();
		playlist_play();
	}
	else
	{
		mainwin_set_info_text();
		playlistwin_update_list();
	}
}

void playlist_prev(void)
{
	GList *plist_pos_list;
	
	PL_LOCK();
	if (!playlist)
	{
		PL_UNLOCK();
		return;
	}

	plist_pos_list = find_playlist_position_list();

	if (!cfg.repeat && !g_list_previous(plist_pos_list))
	{
		PL_UNLOCK();
		return;
	}

	
	if (g_list_previous(plist_pos_list))
		playlist_position = g_list_previous(plist_pos_list)->data;
	PL_UNLOCK();
	playlist_check_pos_current();
	if (get_input_playing())
	{
		input_stop();
		playlist_play();
	}
	else
	{
		mainwin_set_info_text();
		playlistwin_update_list();
	}
}

void playlist_set_position(gint pos)
{
	GList *node;
	PL_LOCK();
	if (!playlist)
	{
		PL_UNLOCK();
		return;
	}

	node = g_list_nth(playlist, pos);
	if(!node)
	{
		PL_UNLOCK();
		return;
	}
	
	playlist_position = node->data;
	PL_UNLOCK();
	playlist_check_pos_current();
	if (get_input_playing())
	{
		input_stop();
		playlist_play();
	}
	else
	{
		mainwin_set_info_text();
		playlistwin_update_list();
	}
}

void playlist_eof_reached(void)
{
	GList *plist_pos_list;

	input_stop();

	PL_LOCK();
	plist_pos_list = find_playlist_position_list();

	if (cfg.no_playlist_advance)
	{
		PL_UNLOCK();
		mainwin_set_song_info(0, 0, 0);
		if (cfg.repeat)
			playlist_play();
		return;
	}
	if (!g_list_next(plist_pos_list))
	{
		if (cfg.shuffle)
		{
			PL_UNLOCK();
  			playlist_generate_shuffle_list();
			PL_LOCK();
			playlist_position = shuffle_list->data;
		}
		else
			playlist_position = playlist->data;
		if (!cfg.repeat)
		{
			PL_UNLOCK();
			mainwin_set_song_info(0, 0, 0);
			mainwin_set_info_text();
			return;
		}
	}
	else
		playlist_position = g_list_next(plist_pos_list)->data;
	PL_UNLOCK();
	playlist_check_pos_current();
	playlist_play();
	mainwin_set_info_text();
	playlistwin_update_list();
}

gint get_playlist_length(void)
{
	gint retval;

	PL_LOCK();
	retval = __get_playlist_length();
	PL_UNLOCK();
	
	return retval;
}

gint __get_playlist_length(void)
{
	/* Caller should hold playlist_mutex */
	if (playlist)
		return(g_list_length(playlist));
	return 0;
}

gchar *playlist_get_info_text(void)
{
	PlaylistEntry *entry;
	gchar *text = NULL, *title, *tmp, *tmp2;

	PL_LOCK();
	if (playlist)
	{
		entry = (PlaylistEntry *) g_list_nth(playlist, get_playlist_position())->data;
		if (entry->title)
			title = entry->title;
		else
		{
			title = strrchr(entry->filename, '/');
			if (!title)
				title = entry->filename;
			else
				title++;
		}
		text = g_malloc(strlen(title) + 20);
		if (entry->length != -1)
			sprintf(text, "%d. %s (%d:%-2.2d)", get_playlist_position() + 1, title, entry->length / 60000, (entry->length / 1000) % 60);
		else
			sprintf(text, "%d. %s", get_playlist_position() + 1, title);
		if (cfg.convert_underscore)
			while ((tmp = strchr(text, '_')) != NULL)
				*tmp = ' ';
		if (cfg.convert_twenty)
			while ((tmp = strstr(text, "%20")) != NULL)
			{
				tmp2 = tmp + 3;
				*(tmp++) = ' ';
				while (*tmp2)
					*(tmp++) = *(tmp2++);
				*tmp = '\0';
			}

	}
	PL_UNLOCK();
	return text;
}

int playlist_get_current_length(void)
{
	gint retval = 0;

	PL_LOCK();
	if (playlist)
		retval = playlist_position->length;
	PL_UNLOCK();
	return retval;
}

gboolean playlist_save(gchar * filename)
{
	PlaylistEntry *entry;
	GList *node;
	FILE *file;
	gchar *ext;
	gboolean is_pls = FALSE;

	if ((file = fopen(filename, "w")) != NULL)
	{
		ext = strrchr(filename, '.');
		if (ext)
		{
			if (!strcasecmp(ext, ".pls"))
			{
				is_pls = TRUE;
				fprintf(file, "[playlist]\n");
				fprintf(file, "NumberOfEntries=%d\n", get_playlist_length());
			}
		}
		PL_LOCK();
		node = playlist;
		while (node)
		{
			entry = (PlaylistEntry *) node->data;
			if (is_pls)
				fprintf(file, "File%d=%s\n", g_list_position(playlist, node) + 1, entry->filename);
			else
				fprintf(file, "%s\n", entry->filename);
			node = node->next;
		}
		PL_UNLOCK();
		fclose(file);
		return TRUE;
	}
	return FALSE;
}

gboolean playlist_load(gchar * filename)
{
	return playlist_load_ins(filename, -1);
}

static void playlist_load_ins_file(gchar * filename, gchar * playlist_name, guint pos)
{
	gint i;
	gchar *temp, *path;

	filename = g_strstrip(filename);

	if(cfg.use_backslash_as_dir_delimiter)
	{
		while(temp = strchr(filename, '\\'))
		      *temp = '/';
	}

	if (filename[0] != '/' && !strstr(filename, "://"))
	{
		path = g_strdup(playlist_name);
		temp = strrchr(path, '/');
		if (temp)
			*temp = '\0';
		else
		{
			__playlist_ins(filename, pos);
			return;
		}
		temp = g_strdup_printf("%s/%s", path, filename);
		__playlist_ins(temp, pos);
		g_free(temp);
		g_free(path);
	}
	else
		__playlist_ins(filename, pos);
}

guint playlist_load_ins(gchar * filename, glong pos)
{
	FILE *file;
	gchar *line, *ext, key[10];
	gint i, noe;
	guint entries = 0;
	int linelen = 1024;

	ext = strrchr(filename, '.');
	if (ext && !strcasecmp(ext, ".pls"))
	{
		if ((line = read_ini_string(filename, "playlist", "NumberOfEntries")) != NULL)
		{
			noe = atoi(line);
			g_free(line);
		}
		else
			return FALSE;
		for (i = 1; i <= noe; i++)
		{
			g_snprintf(key, 10, "File%d", i);
			if ((line = read_ini_string(filename, "playlist", key)) != NULL)
			{
				playlist_load_ins_file(line, filename, pos);
				entries ++;
				if (pos > 0)
					pos++;
				g_free(line);
			}
		}
		playlist_generate_shuffle_list();
		return entries;
	}
	else
	{
		if ((file = fopen(filename, "r")) != NULL)
		{
			line = g_malloc(linelen);
			while (fgets(line, linelen, file))
			{
				while (strlen(line) == linelen - 1 && line[strlen(line) - 1] == '\n')
				{
					linelen += 1024;
					line = (gchar *) g_realloc(line, linelen);
					fgets(&line[strlen(line)], 1024, file);
				}
				while (line[strlen(line) - 1] == '\r' || line[strlen(line) - 1] == '\n')
					line[strlen(line) - 1] = '\0';
				playlist_load_ins_file(line, filename, pos);
				entries++;
				if (pos > 0)
					pos++;
			}
			fclose(file);
			playlist_generate_shuffle_list();
			return entries;
		}
	}
	return FALSE;
}

GList *get_playlist(void)
{
	return playlist;
}

gint get_playlist_position(void)
{
	/* Caller should hold playlist_mutex */
	if (playlist && playlist_position)
		return g_list_index(playlist, playlist_position);
	return 0;
}

gint playlist_sort_by_title_cmpfunc(PlaylistEntry * a, PlaylistEntry * b)
{
	gchar *a_title = NULL, *b_title = NULL;

	if (a->title)
		a_title = a->title;
	else
	{
		if (strrchr(a->filename, '/'))
			a_title = strrchr(a->filename, '/') + 1;
		else
			a_title = a->filename;
	}

	if (b->title)
		b_title = b->title;
	else
	{
		if (strrchr(a->filename, '/'))
			b_title = strrchr(b->filename, '/') + 1;
		else
			b_title = b->filename;

	}
	return strcasecmp(a_title, b_title);
}

void playlist_sort_by_title(void)
{
	PL_LOCK();
	playlist = g_list_sort(playlist, (GCompareFunc) playlist_sort_by_title_cmpfunc);
	PL_UNLOCK();
}

gint playlist_sort_by_filename_cmpfunc(PlaylistEntry * a, PlaylistEntry * b)
{
	gchar *a_filename = NULL, *b_filename = NULL;

	if (strrchr(a->filename, '/'))
		a_filename = strrchr(a->filename, '/') + 1;
	else
		a_filename = a->filename;

	if (strrchr(b->filename,'/'))
		b_filename = strrchr(b->filename, '/') + 1;
	else
		b_filename = b->filename;

	return strcasecmp(a_filename, b_filename);
}

void playlist_sort_by_filename(void)
{
	PL_LOCK();
	playlist = g_list_sort(playlist, (GCompareFunc) playlist_sort_by_filename_cmpfunc);
	PL_UNLOCK();
}

gint playlist_sort_by_path_cmpfunc(PlaylistEntry * a, PlaylistEntry * b)
{
	return strcasecmp(a->filename, b->filename);
}

void playlist_sort_by_path(void)
{
	PL_LOCK();
	playlist = g_list_sort(playlist, (GCompareFunc) playlist_sort_by_path_cmpfunc);
	PL_UNLOCK();
}

static GList* playlist_sort_selected(GList *list, GCompareFunc cmpfunc)
{
	GList *list1, *list2;
	GList *temp_list = NULL;
	GList *index_list = NULL;

	/*
	 * We take all the selected entries out of the playlist,
	 * sorts them, and then put them back in again.
	 */

	list1 = g_list_last(list);

	while(list1)
	{
		list2 = g_list_previous(list1);
		if(((PlaylistEntry *) list1->data)->selected)
		{
			index_list = g_list_prepend(index_list, GINT_TO_POINTER(g_list_position(list, list1)));
			list = g_list_remove_link(list, list1);
			temp_list = g_list_concat(list1, temp_list);
		}
		list1 = list2;
	}

	temp_list = g_list_sort(temp_list, cmpfunc);
	list1 = temp_list;
	list2 = index_list;

	while(list2)
	{
		if(!list1)
		{
			g_log(NULL, G_LOG_LEVEL_CRITICAL, "%s: Error during list sorting\nPossibly dropped some playlist-entries.", PACKAGE);
			break;
		}
		list = g_list_insert(list, list1->data, GPOINTER_TO_INT(list2->data));
		list2 = g_list_next(list2);
		list1 = g_list_next(list1);
	}
	g_list_free(index_list);
	g_list_free(temp_list);

	return list;
}


void playlist_sort_selected_by_title(void)
{
	PL_LOCK();	
	playlist = playlist_sort_selected(playlist, (GCompareFunc) playlist_sort_by_title_cmpfunc);
	PL_UNLOCK();
}

void playlist_sort_selected_by_filename(void)
{
	PL_LOCK();
	playlist = playlist_sort_selected(playlist, (GCompareFunc) playlist_sort_by_filename_cmpfunc);
	PL_UNLOCK();
}

void playlist_sort_selected_by_path(void)
{
	PL_LOCK();
	playlist = playlist_sort_selected(playlist, (GCompareFunc) playlist_sort_by_path_cmpfunc);
	PL_UNLOCK();
}

void playlist_reverse(void)
{
	PL_LOCK();
	playlist = g_list_reverse(playlist);
	PL_UNLOCK();
}

void playlist_random(void)
{
	GList *new_list, *node, *prev;
	gint numsongs;

	PL_LOCK();

	numsongs = g_list_length(playlist);

	new_list = g_list_nth(playlist, (gint) ((gfloat) numsongs-- * rand() / (RAND_MAX + 1.0)));
	playlist = g_list_remove_link(playlist, new_list);
	prev = new_list;

	while (numsongs)
	{
		node = g_list_nth(playlist, (gint) ((gfloat) numsongs-- * rand() / (RAND_MAX + 1.0)));
		playlist = g_list_remove_link(playlist, node);
		prev->next = node;
		node->prev = prev;
		prev = node;
	}

	playlist = new_list;
	PL_UNLOCK();
}

void playlist_generate_shuffle_list(void)
{
	GList *new_list, *node, *prev;
	gint numsongs;
	gboolean update_mainwin = FALSE;

	PL_LOCK();

	if (shuffle_list)
	{
		g_list_free(shuffle_list);
		shuffle_list = NULL;
	}
	if (playlist)
	{
		node = playlist;
		while (node)
		{
			shuffle_list = g_list_append(shuffle_list, node->data);
			node = g_list_next(node);
		}
		numsongs = g_list_length(shuffle_list);

		if (!numsongs)
		{
			PL_UNLOCK();
			return;
		}

		new_list = g_list_nth(shuffle_list, (gint) ((gfloat) numsongs-- * rand() / (RAND_MAX + 1.0)));
		shuffle_list = g_list_remove_link(shuffle_list, new_list);
		prev = new_list;

		while (numsongs)
		{
			node = g_list_nth(shuffle_list, (gint) ((gfloat) numsongs-- * rand() / (RAND_MAX + 1.0)));
			shuffle_list = g_list_remove_link(shuffle_list, node);
			prev->next = node;
			node->prev = prev;
			prev = node;
		}
		shuffle_list = new_list;

		if (!playlist_position)
		{
			if (cfg.shuffle)
				playlist_position = shuffle_list->data;
			else
				playlist_position = playlist->data;
			update_mainwin = TRUE;
		}
	}
	PL_UNLOCK();
	if (update_mainwin)
		mainwin_set_info_text();
}

void *playlist_get_info_func(void *arg)
{
	GList *node, *playlist;
	gboolean update_playlistwin = FALSE, update_mainwin = FALSE;
	PlaylistEntry *entry;
	gchar *temp_filename, *temp_title;
	gint temp_length;

	while (playlist_get_info_going)
	{
		pthread_mutex_lock(&playlist_mutex);
		playlist = get_playlist();
		if (cfg.get_info_on_load && playlist_get_info_scan_active)
		{
			for (node = playlist; node; node = g_list_next(node))
			{
				entry = node->data;
				if (entry->title || entry->length != -1)
					continue;
				
				temp_filename = g_strdup(entry->filename);
				temp_title = NULL;
				temp_length = -1;

				/* We don't want to lock the playlist while reading info */
				PL_UNLOCK();
				input_get_song_info(temp_filename, &temp_title, &temp_length);
				PL_LOCK();
				g_free(temp_filename);
				/* Make sure entry is still around */
				if(g_list_index(playlist, entry) != -1)
				{
					/* entry is still around */
					if (!temp_title && temp_length == -1)
						continue;
					entry->title = temp_title;
					entry->length = temp_length;
					if (playlistwin_item_visible(g_list_position(playlist, node)))
						update_playlistwin = TRUE;
					if (entry == (PlaylistEntry *) g_list_nth(get_playlist(),
										  get_playlist_position())->data)
						update_mainwin = TRUE;
					break;
				}
				else
				{
					/* entry wasn't around anymore so we restart. */
					node = playlist;
					continue;
				}	
			}

			if (!node)
				playlist_get_info_scan_active = FALSE;
		}
		else if (cfg.get_info_on_demand && cfg.playlist_visible && !cfg.playlist_shaded && playlist)
		{
			for (node = g_list_nth(playlist, playlistwin_get_toprow()); node && playlistwin_item_visible(g_list_position(playlist, node)); node = g_list_next(node))
			{
				entry = node->data;
				if (entry->title || entry->length != -1)
					continue;
				
				temp_filename = g_strdup(entry->filename);
				temp_title = NULL;
				temp_length = -1;

				/* We don't want to lock the playlist while reading info */
				PL_UNLOCK();
				input_get_song_info(temp_filename, &temp_title, &temp_length);
				PL_LOCK();
				g_free(temp_filename);
				/* Make sure entry is still around */
				if(g_list_index(playlist, entry) != -1)
				{
					/* entry is still around */
					if (!temp_title && temp_length == -1)
						continue;
					entry->title = temp_title;
					entry->length = temp_length;
					if (playlistwin_item_visible(g_list_position(playlist, node)))
						update_playlistwin = TRUE;
					if (entry == (PlaylistEntry *) g_list_nth(get_playlist(),
										  get_playlist_position())->data)
						update_mainwin = TRUE;
					break;
				}
				else
				{
					/* entry wasn't around anymore so we restart. */
					node = g_list_nth(playlist, playlistwin_get_toprow());
					continue;
				}				
			}
		}
		pthread_mutex_unlock(&playlist_mutex);
		if (update_playlistwin)
		{
			playlistwin_update_list();
			update_playlistwin = FALSE;
		}
		if (update_mainwin)
		{
			mainwin_set_info_text();
			update_mainwin = FALSE;
		}
		xmms_usleep(10000);
	}
	pthread_exit(NULL);
}

void playlist_remove_dead_files(void)
{
}

void playlist_start_get_info_thread(void)
{
	playlist_get_info_going = TRUE;
	pthread_create(&playlist_get_info_thread, NULL, playlist_get_info_func, NULL);
}

void playlist_stop_get_info_thread(void)
{
	playlist_get_info_going = FALSE;
	pthread_join(playlist_get_info_thread, NULL);
}

void playlist_start_get_info_scan(void)
{
	playlist_get_info_scan_active = TRUE;
}
